#include "kernel_impl.h"
#include <sil.h>
#include "mmu.h"

/*
 * ̃x[XAhXƃTCY
 */
#define SIZE_1M    0x00100000

#define DOMAIN_NO  1

#define PAGE_TABLE_SIZE  0x00004000 /* 4KB */

/*
 * ϊe[uւ̐ݒe
 * va   : zAhX
 * pa   : AhX
 * size : TCY 
 * ns   : Non-Secure access
 * s    : Lw
 * tex  : C BƂ̑gݍ킹ŕω
 * ap   : ANZX
 * c    : LbV
 * c    : obt@
 */
typedef struct{
    uintptr_t   va;
    uintptr_t   pa;
    uint32_t    size;
    uint8_t     ns;
    uint8_t     s;
    uint8_t     tex;
    uint8_t     ap;
    uint8_t     c;
    uint8_t     b;
}MEMORY_ATTRIBUTE;

/*
 * y[We[u
 * TTBCR  N  0 w肵Ă邽, 16KB EłȂ΂ȂȂ.
 */
uint32_t PageTable[PAGE_TABLE_SIZE] __attribute__((aligned(16384)));

/*
 * y[We[ũZNVݒ肷
 */
static void
mmu_set_section(uintptr_t va, uintptr_t pa,
                uint8_t ns, uint8_t s, uint8_t tex,
                uint8_t ap, uint8_t c, uint8_t b)
{
    uint32_t ttb;
    uint32_t *sptr;

    /*
     * y[We[ũx[XAhX擾
     */
    CP15_TTB0_READ(ttb);

    /*
     * VAZNṼAhX擾
     */
    sptr  = (uint32_t *)((ttb & 0xFFFFC000) | (((va & 0xFFF00000) >> 20) << 2));
    *sptr = pa | (ns << 19) | (s << 16) | (tex << 12) | (ap << 10) | (c << 3) | (b << 2) | (1 << 1);
}

/*
 * wɏ],̃}bsOs
 * }bsO̒Pʂ1MB 
 */
static void
mmu_map_memory(MEMORY_ATTRIBUTE *m_attribute)
{
    uint32_t size;
    uintptr_t va;
    uintptr_t pa;
    
    size = m_attribute->size;
    va   = m_attribute->va;
    pa   = m_attribute->pa;

    /* MBPʂɂȂ悤ɃTCYݒ */
    size = (size + 0x000FFFFF) & ~0x000FFFFF;

    while(size > 0){
        mmu_set_section(va, pa,
                        m_attribute->ns,
                        m_attribute->s,
                        m_attribute->tex,
                        m_attribute->ap,
                        m_attribute->c,
                        m_attribute->b);

        va   += SIZE_1M;
        pa   += SIZE_1M;
        size -= SIZE_1M;
    }
}

void
pb1176_mmu_init(void){
    MEMORY_ATTRIBUTE m_attribute;

    uint32_t bits = 0;

    /* vtFb`obt@NA */
    pb1176_pbuffer_flash();

    /*
     * TTBR0ppɎw
     */
    CP15_TTBCR_WRITE(0);

    /*
     * ϊe[u(TT)mۂ
     * Crbg(Inner Cachable)Zbg 
     */
    CP15_TTB0_WRITE((((uint32_t)PageTable)|CP15_TTB0_C));

    /* vtFb`obt@NA */
    pb1176_pbuffer_flash();

    /*
     * SZNV𕨗AhX = zAhX
     * mLbVuCmobt@uƂ 
     */
    m_attribute.pa   = 0x00000000;
    m_attribute.va   = m_attribute.pa;
    m_attribute.size = 0x80000000;
    m_attribute.ns   = 0;          /* ns=0: Secure access */
    m_attribute.s    = 0;          /* not-shared */
    m_attribute.ap   = 3;          /* tANZX */
    m_attribute.tex  = 0;          /* 000 00 Strongly Ordered */
    m_attribute.c    = 0;
    m_attribute.b    = 0;

    mmu_map_memory (&m_attribute);

    /*
     * SRAM/RAM̐ݒ
     */
#ifdef TOPPERS_WITH_SAFEG
    m_attribute.pa   = 0x10300000;
#else
    m_attribute.pa   = 0x0;
#endif

    m_attribute.va   = m_attribute.pa;
    m_attribute.size = SIZE_1M*2;
    m_attribute.c    = 1;    /* cb:11 No Allocate on Write */
    m_attribute.b    = 1;

    mmu_map_memory (&m_attribute);

    /*
     * Set 1MB of shared memory between both worlds
     */
#ifdef TOPPERS_WITH_SAFEG
    m_attribute.pa   = 0x4000000; // 0x6000000 + SIZE_1M*2; // 0x6200000
    m_attribute.va   = m_attribute.pa;
    m_attribute.size = SIZE_1M;
    m_attribute.ns   = 1;    /* 1: Non-Secure access */
    m_attribute.s    = 1;    /* 1:shared */
    m_attribute.ap   = 3;    /* tANZX */
    m_attribute.tex  = 1;    /* 100 Outer and Inner Noncachable */
    m_attribute.c    = 0;
    m_attribute.b    = 0;
#endif

    /*
     * TCM settings
     */
    m_attribute.pa   = TCM_BASE_ADDRESS;
    m_attribute.va   = m_attribute.pa;
    m_attribute.size = SIZE_1M;
    m_attribute.ns   = 1;    /* ns=1: Non-Secure access */
    m_attribute.s    = 0;    /* not-shared */
    m_attribute.ap   = 3;    /* tANZX */
    m_attribute.tex  = 1;    /* Outer and Inner Noncachable */
    m_attribute.c    = 0;
    m_attribute.b    = 0;

    mmu_map_memory (&m_attribute);

    /*
     * hCԍZbg
     */
    CP15_DOMAINS_WRITE(DOMAIN_NO);

    /*
     * CONTROLRvZbT MrbgCXPrbgZbgāC
     * MMULɂ
     */
    CP15_CONTROL_READ(bits);
    bits |= CP15_CONTROL_M_BIT | CP15_CONTROL_XP_BIT;
    CP15_CONTROL_WRITE(bits);
}

/*
 * icache L
 * eratta ɂ悤, i cache  invalidate 1xŌȂƂ̂,
 * sKv.
 */
void
icache_enable(void)
{
    unsigned long int bits = 0;
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_ICACHE_INVALIDATE();
    CP15_CONTROL_READ(bits);
    bits |= CP15_CONTROL_I_BIT;
    CP15_CONTROL_WRITE(bits);
}

/* dcache L */
void
dcache_enable(void)
{
    unsigned long int bits = 0;
    CP15_DCACHE_CLEAN_AND_INVALIDATE();
    CP15_CONTROL_READ(bits);
    bits |= CP15_CONTROL_C_BIT;
    CP15_CONTROL_WRITE(bits);
}

void pb1176_cache_init(){
    dcache_enable();
    icache_enable();
}
